@@ -705,6 +705,10 @@ back to the user from the mailer."
 		(save-excursion
 		  (set-buffer errbuf)
 		  (erase-buffer))))
+
+	  ;; XEmacs change
+	  (mail-check-safe-charset)
+
 	  (let ((default-directory "/")
 		exit-value)
 	    (setq exit-value (apply 'call-process-region
